import java.io.* ;
import java.net.* ;

class Message
{

    DatagramPacket packet ; 

    final static int MAX_LENGTH = 512 ;

    // Types of messages
    final static short PING = 1 ;
    final static short PING_REP = 2 ;
    final static short JOIN = 3 ;
    final static short PUBLISH = 4 ;
    final static short GET = 5 ;
    final static short GET_REP = 6 ;

    // TTL for throwing old messages
    final static short TTL_INIT = 255 ;
    final static short TTL_ONE = 1 ;

    //     Message format :
    // type (2 bytes)
    // TTL (2 bytes)
    // contentID (8 bytes)
    // fromID (8 bytes)
    // fromIP (4 bytes)
    // fromPort (2 bytes)
    // reserved for futur use (2 bytes), should be 0
    // succID (8 bytes)
    // succIP (4 bytes)
    // succPort (2 bytes)
    // dataLength (2 bytes)
    // data (dataLength bytes)


// Constructors allow to create new messages of each type -----------------

    Message () // blank message for reception
    {
	packet = new DatagramPacket (new byte[MAX_LENGTH], MAX_LENGTH) ;
    }

    // for sending a message :
    Message (short type, short ttl, long contentID, PeerInfo from, PeerInfo succ)  
    {
	packet = new DatagramPacket (new byte[MAX_LENGTH], MAX_LENGTH) ;
	setType (type) ;
	setTTL (ttl) ;
	setContentID (contentID) ;
	setFromID (from.id) ;
	setFromIP (from.addr) ;
	setFromPort (from.port) ;
	setSuccID (succ.id) ;
	setSuccIP (succ.addr) ;
	setSuccPort (succ.port) ;
    }

    public String toString ()
    {
	String udp = "UDP: " + packet.getAddress () + ":" + packet.getPort () + "\n" ;
	return udp + BytesTools.bytesToString (packet.getData(), 0, length ()) ;
    }

    int length ()
    {
	int length = 44 ;
	switch (getType ()) {
	case PUBLISH: case GET_REP: length = length + getDataLength() ; break ;
	}
	return length ;
    }

// Functions for accessing messages parts ----------------------------------


    PeerInfo getFromPeer ()
    {
	return new PeerInfo (getFromID(), getFromIP(), getFromPort()) ;
    }

    PeerInfo getSuccPeer ()
    {
	return new PeerInfo (getSuccID(), getSuccIP(), getSuccPort()) ;
    }

    short getType ()
    {
	return BytesTools.bytesToShort (packet.getData(), 0) ;
    }

    short getTTL ()
    {
	return BytesTools.bytesToShort (packet.getData(), 2) ;
    }

    long getContentID ()
    {
	return BytesTools.bytesToLong (packet.getData(), 4) ;
    }

    long getFromID ()
    {
	return BytesTools.bytesToLong (packet.getData(), 12) ;
    }

    InetAddress getFromIP ()
    {
	return BytesTools.bytesToInetAddress (packet.getData(), 20) ;
    }

    short getFromPort ()
    {
	return BytesTools.bytesToShort (packet.getData(), 24) ;
    }

    long getSuccID ()
    {
	return BytesTools.bytesToLong (packet.getData(), 28) ;
    }

    InetAddress getSuccIP ()
    {
	return BytesTools.bytesToInetAddress (packet.getData(), 36) ;
    }

    short getSuccPort ()
    {
	return BytesTools.bytesToShort (packet.getData(), 40) ;
    }

    short getDataLength ()
    {
	return BytesTools.bytesToShort (packet.getData(), 42) ;
    }

    String getData () 
    {
	try {
	    return new String (packet.getData(), 44, getDataLength(), "8859_1") ;
	} catch (UnsupportedEncodingException e) {
	    return new String (packet.getData(), 44, getDataLength()) ;
	}
    }


    void setType (short t)
    {
	BytesTools.shortToBytes (t, packet.getData(), 0) ;
    }

    void setTTL (short t)
    {
	BytesTools.shortToBytes (t, packet.getData(), 2) ;
    }

    void setContentID (long l)
    {
	BytesTools.longToBytes (l, packet.getData(), 4) ;
    }

    void setFromID (long l)
    {
	BytesTools.longToBytes (l, packet.getData(), 12) ;
    }

    void setFromIP (InetAddress a)
    {
	BytesTools.inetAddressToBytes (a, packet.getData(), 20) ;
    }

    void setFromPort (short p)
    {
	BytesTools.shortToBytes (p, packet.getData(), 24) ;
    }

    void setSuccID (long l)
    {
	BytesTools.longToBytes (l, packet.getData(), 28) ;
    }

    void setSuccIP (InetAddress a)
    {
	BytesTools.inetAddressToBytes (a, packet.getData(), 36) ;
    }

    void setSuccPort (short p)
    {
	BytesTools.shortToBytes (p, packet.getData(), 40) ;
    }

    void setDataLength (short len) 
    {
	BytesTools.shortToBytes (len, packet.getData(), 42) ;
    }

    void setData (String s) 
    {
	setDataLength ((short)s.length()) ;
	try {
	    BytesTools.bytesToBytes (s.getBytes("8859_1"), 0, s.length(), packet.getData(), 44) ;
	} catch (UnsupportedEncodingException e) {
	    BytesTools.bytesToBytes (s.getBytes(), 0, s.length(), packet.getData(), 44) ;
	}
    }

}

class BytesTools
{
    // Some tools for coding and decoding short, int, long, InetAddress
    // and String in bytes.
    //  (high order bytes first)

    // store s as 2 bytes in to at position pos
    static void shortToBytes (short s, byte[] to, int pos)
    {	
	byte high = (byte)((s >> 8) & 0xFF) ;
	byte low = (byte)(s & 0xFF) ;
	to [pos] = high ;
	to [pos + 1] = low ;
    }
    
    // read a short at position pos in from
    static short bytesToShort (byte[] from, int pos)
    {
	int high = from[pos] & 0xFF ;
	int low = from[pos + 1] & 0xFF ;
	return (short)((high << 8) | low) ;
    }
    
    // store i as 4 bytes in to at position pos
    static void intToBytes (int i, byte[] to, int pos)
    {
	short high = (short)((i >> 16) & 0xFFFF) ;
	short low = (short)(i & 0xFFFF) ;
	shortToBytes (high, to, pos) ;
	shortToBytes (low, to, pos + 2) ;
    }

    // read an int at position pos in from
    static int bytesToInt (byte[] from, int pos)
    {
	int high = bytesToShort (from, pos) & 0xFFFF ;
	int low =  bytesToShort (from, pos + 2) & 0xFFFF ;
	return (high << 16) | low ;
    }

    // store l as 8 bytes in to at position pos
    static void longToBytes (long l, byte[] to, int pos)
    {
	int high = (int)((l >> 32) & 0xFFFFFFFF) ;
	int low = (int)(l & 0xFFFFFFFF) ;
	intToBytes (high, to, pos) ;
	intToBytes (low, to, pos + 4) ;
    }

    // read an long at position pos in from
    static long bytesToLong (byte[] from, int pos)
    {
	long high = bytesToInt (from, pos) & 0xFFFFFFFFL ;
	long low =  bytesToInt (from, pos + 4) & 0xFFFFFFFFL ;
	return (high << 32) | low ;
    }

    // return a printable string of buf[from]..buf[from+len-1]
    static String bytesToString (byte[] buf, int from, int len)
    {
	StringBuffer s = new StringBuffer () ;
	int relPos = 0 ;
	while (len-- > 0) {
	    int i = ((int) buf[from++]) & 0xFF ;
	    s.append (i/100) ; i = i % 100 ;
	    s.append (i/10) ; i = i % 10 ;
	    s.append (i) ;

	    if (relPos++ % 4 == 3) s.append ('\n') ;
	    else s.append ('.') ;
	}
	if (relPos % 4 != 0) s.append ('\n') ;
	return s.toString () ;
    }

    // Read a raw ip address at pos from in buf
    static InetAddress bytesToInetAddress (byte[] buf, int from)
    {
	try {
	    String s = bytesToString (buf, from, 4) ;
	    String ip = s.substring (0, s.length() - 1) ;
	    return InetAddress.getByName (ip) ;
	} catch (UnknownHostException e) {
	    return null ;
	}
    }

    // Copy len bytes at position fromSrc in src to dst at position fromDst
    static void bytesToBytes (byte[] src, int fromSrc, int len, byte[] dst, int fromDst)
    {
	while (len-- > 0) {
	    dst [fromDst++] = src [fromSrc++] ;
	}
    } 

    // Write an ip address at position pos in to
    static void inetAddressToBytes (InetAddress a, byte[] to, int pos)
    {
	byte[] raw = a.getAddress () ;
	bytesToBytes (raw, 0, 4, to, pos) ;
    }

}

